{Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved.}

{TERATERM.EXE, scroll buffer routines}
unit Buffer;

interface
{$I teraterm.inc}

{$IFDEF Delphi}
uses WinTypes, WinProcs, OWindows, TTTypes;
{$ELSE}
uses WinTypes, WinProcs, WObjects, TTTypes;
{$ENDIF}

const
  BuffXMax = 300;
  BuffYMax = 800;
  BuffSizeMax: longint = 65535;

var
  HCodeBuff, HAttrBuff, HAttrBuff2: THandle;
  BuffLock: boolean;
  BufferSize: LongInt;
  PageStart: integer;
  BuffStartAbs, BuffEnd, BuffEndAbs: integer;
  BufferFull: boolean;

  CodeBuff: PChar;  {Character code buffer}
  AttrBuff: PChar;  {Attribute buffer}
  AttrBuff2: PChar; {Color attr buffer}
  LinePtr: longint;
  CodeLine: PChar;
  AttrLine: PChar;
  AttrLine2: PChar;

  NumOfLines, NumOfColumns: integer;
  NumOfLinesInBuff: integer;

  Select: boolean;
  SelectStart, SelectEnd, SelectEndOld: TPoint;

function ChangeBuffer(Nx, Ny: integer): boolean;
procedure LockBuffer;
procedure UnlockBuffer;
procedure FreeBuffer;
procedure ResetBuff;
function GetLinePtr(Line: integer): longint;
function NextLinePtr(Ptr: longint): longint;
function PrevLinePtr(Ptr: longint): longint;
procedure NewLine(Line: integer);
procedure NextLine;
procedure PrevLine;
procedure BuffScroll(Count: integer);

implementation

function ChangeBuffer(Nx, Ny: integer): boolean;
var
  HCodeNew, HAttrNew, HAttr2New: THandle;
  NewSize: longint;
  NxCopy, NyCopy, i: integer;
  CodeDest, AttrDest, AttrDest2: PChar;
  Ptr: pointer;
  SrcPtr, DestPtr: longint;
  LockOld: boolean;
begin
  ChangeBuffer := false;

  if Nx > BuffXMax then Nx := BuffXMax;
  if Ny > BuffYMax then Ny := BuffYMax;
 
  if longmul(Nx,Ny) > BuffSizeMax then
    Ny := BuffSizeMax div Nx;

  NewSize := longmul(Nx,Ny);

  HCodeNew := GlobalAlloc(gmem_moveable, NewSize);
  if HCodeNew=0 then exit;
  Ptr := GlobalLock(HCodeNew);
  if Ptr=nil then
  begin
    GlobalFree(HCodeNew);
    exit;
  end;
  CodeDest := PChar(Ptr);

  HAttrNew := GlobalAlloc(gmem_moveable, NewSize);
  if HAttrNew=0 then
  begin
    GlobalFree(HCodeNew);
    exit;
  end;
  Ptr := GlobalLock(HAttrNew);
  if Ptr=nil then
  begin
    GlobalFree(HCodeNew);
    GlobalFree(HAttrNew);
    exit;
  end;
  AttrDest := PChar(Ptr);

  HAttr2New := GlobalAlloc(gmem_moveable, NewSize);
  if HAttr2New=0 then
  begin
    GlobalFree(HCodeNew);
    GlobalFree(HAttrNew);
    exit;
  end;
  Ptr := GlobalLock(HAttr2New);
  if Ptr=nil then
  begin
    GlobalFree(HCodeNew);
    GlobalFree(HAttrNew);
    GlobalFree(HAttr2New);
    exit;
  end;
  AttrDest2 := PChar(Ptr);

  FillChar(CodeDest[0], word(NewSize), $20);
  FillChar(AttrDest[0], word(NewSize), AttrDefault);
  FillChar(AttrDest2[0], word(NewSize), AttrDefault2);

  if HCodeBuff<>0 then
  begin
    if NumOfColumns > Nx then NxCopy := Nx
                         else NxCopy := NumOfColumns;
    if BuffEnd > Ny then NyCopy := Ny
                    else NyCopy := BuffEnd;
    LockOld := BuffLock;
    LockBuffer;
    SrcPtr := GetLinePtr(BuffEnd-NyCopy);
    DestPtr := 0;
    for i := 1 to NyCopy do
    begin
      Move(CodeBuff[SrcPtr], CodeDest[DestPtr], NxCopy);
      Move(AttrBuff[SrcPtr], AttrDest[DestPtr], NxCopy);
      Move(AttrBuff2[SrcPtr], AttrDest2[DestPtr], NxCopy);
      SrcPtr := NextLinePtr(SrcPtr);
      DestPtr := DestPtr + longint(Nx);
    end;
    FreeBuffer;
  end
  else begin
    LockOld := false;
    NyCopy := NumOfLines;
    Select := false;
  end;

  if Select then
  begin
    SelectStart.Y := SelectStart.Y - BuffEnd + NyCopy;
    SelectEnd.Y := SelectEnd.Y - BuffEnd + NyCopy;
    if SelectStart.Y < 0 then
    begin
      SelectStart.Y := 0;
      SelectStart.X := 0;
    end;
    if SelectEnd.Y<0 then
    begin
      SelectEnd.X := 0;
      SelectEnd.Y := 0;
    end;

    Select := (SelectEnd.Y > SelectStart.Y) or
              ((SelectEnd.Y=SelectStart.Y) and
               (SelectEnd.X > SelectStart.X));
  end;

  HCodeBuff := HCodeNew;
  HAttrBuff := HAttrNew;
  HAttrBuff2 := HAttr2New;
  BufferSize := NewSize;
  NumOfLinesInBuff := Ny;
  BuffStartAbs := 0;
  BuffEnd := NyCopy;

  if BuffEnd=NumOfLinesInBuff then
  begin
    BufferFull := true;
    BuffEndAbs := 0;
  end
  else begin
    BufferFull := false;
    BuffEndAbs := BuffEnd;
  end;
  PageStart := BuffEnd - NumOfLines;

  LinePtr := 0;
  if LockOld then
  begin
    CodeBuff := PChar(GlobalLock(HCodeBuff));
    AttrBuff := PChar(GlobalLock(HAttrBuff));
    AttrBuff2 := PChar(GlobalLock(HAttrBuff2));
    CodeLine := CodeBuff;
    AttrLine := AttrBuff;
    AttrLine2 := AttrBuff2;
  end
  else begin
    GlobalUnlock(HCodeNew);
    GlobalUnlock(HAttrNew);
  end;
  BuffLock := LockOld;

  ChangeBuffer := true;
end;

procedure LockBuffer;
begin
  if BuffLock then exit;
  CodeBuff := PChar(GlobalLock(HCodeBuff));
  AttrBuff := PChar(GlobalLock(HAttrBuff));
  AttrBuff2 := PChar(GlobalLock(HAttrBuff2));
  CodeLine := @CodeBuff[LinePtr];
  AttrLine := @AttrBuff[LinePtr];
  AttrLine2 := @AttrBuff2[LinePtr];
  BuffLock := true;
end;

procedure UnlockBuffer;
begin
  if not BuffLock then exit;
  GlobalUnlock(HCodeBuff);
  GlobalUnlock(HAttrBuff);
  GlobalUnlock(HAttrBuff2);
  BuffLock := false;
end;

procedure FreeBuffer;
begin
  UnlockBuffer;
  GlobalFree(HCodeBuff);
  GlobalFree(HAttrBuff);
  GlobalFree(HAttrBuff2);
end;

procedure ResetBuff;
begin
  PageStart := 0;
  BuffStartAbs := 0;
  BuffEnd := NumOfLines;
  if NumOfLines=NumOfLinesInBuff then
  begin
    BufferFull := true;
    BuffEndAbs := 0;
  end
  else begin
    BufferFull := false;
    BuffEndAbs := NumOfLines;
  end;

  SelectStart.X := 0;
  SelectStart.Y := 0;
  SelectEnd := SelectStart;
  SelectEndOld := SelectStart;
  Select := false;

  NewLine(0);
  FillChar(CodeBuff[0], BufferSize, $20);
  FillChar(AttrBuff[0], BufferSize, AttrDefault);
  FillChar(AttrBuff2[0], BufferSize, AttrDefault2);
end;

function GetLinePtr(Line: integer): longint;
var
  Ptr: longint;
begin
  Ptr := longmul(BuffStartAbs + Line, NumOfColumns);
  if Ptr >= BufferSize then Ptr := Ptr - BufferSize;
  GetLinePtr := Ptr;
end;

function NextLinePtr(Ptr: longint): longint;
begin
  Ptr := Ptr + longint(NumOfColumns);
  if Ptr >= BufferSize then Ptr := Ptr - BufferSize; 
  NextLinePtr := Ptr;
end;

function PrevLinePtr(Ptr: longint): longint;
begin
  Ptr := Ptr - longint(NumOfColumns);
  if Ptr < 0 then Ptr := Ptr + BufferSize;
  PrevLinePtr := Ptr;
end;

procedure NewLine(Line: integer);
begin
  LinePtr := GetLinePtr(Line);
  CodeLine := @CodeBuff[LinePtr];
  AttrLine := @AttrBuff[LinePtr];
  AttrLine2 := @AttrBuff2[LinePtr];
end;

procedure NextLine;
begin
  LinePtr := NextLinePtr(LinePtr);
  CodeLine := @CodeBuff[LinePtr];
  AttrLine := @AttrBuff[LinePtr];
  AttrLine2 := @AttrBuff2[LinePtr];
end;

procedure PrevLine;
begin
  LinePtr := PrevLinePtr(LinePtr);
  CodeLine := @CodeBuff[LinePtr];
  AttrLine := @AttrBuff[LinePtr];
  AttrLine2 := @AttrBuff2[LinePtr];
end;

procedure BuffScroll(Count: integer);
var
  i: integer;
  TmpPtr: longint;
begin
  TmpPtr := GetLinePtr(BuffEnd);
  for i := 1 to Count do
  begin
    FillChar(CodeBuff[TmpPtr],NumOfColumns,$20);
    FillChar(AttrBuff[TmpPtr],NumOfColumns,AttrDefault);
    FillChar(AttrBuff2[TmpPtr],NumOfColumns,AttrDefault2);
    TmpPtr := NextLinePtr(TmpPtr);
    inc(BuffEndAbs);
    if BuffEndAbs >= NumOfLinesInBuff then BuffEndAbs := 0;
    if BufferFull then
    begin
      BuffStartAbs := BuffEndAbs;
      if Select then
      begin
        Dec(SelectStart.Y);
        Dec(SelectEnd.Y);
        if SelectStart.Y<0 then
        begin
          SelectStart.X := 0;
          SelectStart.Y := 0;
        end;
        if SelectEnd.Y<0 then
        begin
          SelectEnd.X := 0;
          SelectEnd.Y := 0;
        end;
        Select := (SelectEnd.Y > SelectStart.Y) or
                  ((SelectEnd.Y=SelectStart.Y) and
                   (SelectEnd.X > SelectStart.X));
      end;
    end
    else begin
      Inc(PageStart);
      Inc(BuffEnd);
      BufferFull := BuffEnd = NumOfLinesInBuff;
    end;
    LinePtr := NextLinePtr(LinePtr);
  end;

  CodeLine := @CodeBuff[LinePtr];
  AttrLine := @AttrBuff[LinePtr];
  AttrLine2 := @AttrBuff2[LinePtr];
end;


begin
  HCodeBuff := 0;
  HAttrBuff := 0;
  HAttrBuff2 := 0;
  BuffLock := false;
end.