{Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved.}

{TERATERM.EXE, TELNET routines}
unit Telnet;

interface

uses WinTypes, WinProcs, Strings, TTCommon, TTWinMan, CommLib;

const
  TEL_EOF    = 236;
  SUSP       = 237;
  ABORT      = 238;

  SE	     = 240;
  NOP	     = 241;
  DM	     = 242;
  BREAK	     = 243;
  IP	     = 244;
  AO	     = 245;
  AYT	     = 246;
  EC	     = 247;
  EL	     = 248;
  GOAHEAD    = 249;
  SB	     = 250;
  WILLTEL    = 251;
  WONTTEL    = 252;
  DOTEL	     = 253;
  DONTTEL    = 254;
  IAC	     = 255;

  BINARY     = 0;
  ECHO	     = 1;
  RECONNECT  = 2;
  SGA 	     = 3;
  AMSN	     = 4;
  STATUS     = 5;
  TIMING     = 6;
  RCTAN	     = 7;
  OLW	     = 8;
  OPS	     = 9;
  OCRD	     = 10;
  OHTS	     = 11;
  OHTD	     = 12;
  OFFD	     = 13;
  OVTS	     = 14;
  OVTD	     = 15;
  OLFD	     = 16;
  XASCII     = 17;
  LOGOUT     = 18;
  BYTEM	     = 19;
  DET	     = 20;
  SUPDUP     = 21;
  SUPDUPOUT  = 22;
  SENDLOC    = 23;
  TERMTYPE   = 24;
  EOR	     = 25;
  TACACSUID  = 26;
  OUTPUTMARK = 27;
  TERMLOCNUM = 28;
  REGIME3270 = 29;
  X3PAD	     = 30;
  NAWS	     = 31;
  TERMSPEED  = 32;
  TFLOWCNTRL = 33;
  LINEMODE   = 34;
  MaxTelOpt  = 34;

  {Telnet status}
  TelIdle    = 0;
  TelIAC     = 1;
  TelSB      = 2;
  TelWill    = 3;
  TelWont    = 4;
  TelDo      = 5;
  TelDont    = 6;
  TelNop     = 7;

type
  OptStatus = (No, Yes, WantNo, WantYes);
  OptQue = (Empty, Opposite);

  PTelOpt = ^TelOpt;
  TelOpt = record
    Accept: boolean;
    Status: OptStatus;
    Que: OptQue;
  end;

  PTelRec = ^TelRec;
  TelRec = record
    MyOpt: array[0..MaxTelOpt] of TelOpt;
    HisOpt: array[0..MaxTelOpt] of TelOpt;
    SubOptBuff: array[0..50] of byte;
    SubOptCount: integer;
    SubOptIAC: boolean;
    ChangeWinSize: boolean;
    WinSize: TPoint;
    LogFile: integer;
  end;

  procedure InitTelnet;
  procedure DefaultTelRec;
  procedure EndTelnet;
  procedure ParseTel;
  procedure TelEnableHisOpt(b: byte);
  procedure TelDisableHisOpt(b: byte);
  procedure TelEnableMyOpt(b: byte);
  procedure TelDisableMyOpt(b: byte);
  procedure TelInformWinSize;
  procedure TelSendAYT;
  procedure TelSendBreak;

var
  TelnetRec: TelRec;
  TelStatus: integer;

implementation

procedure InitTelnet;
begin
  TelStatus := TelIdle;

  DefaultTelRec;
  TelnetRec.MyOpt[BINARY].Accept := true;
  TelnetRec.HisOpt[BINARY].Accept := true;
  TelnetRec.MyOpt[SGA].Accept := true;
  TelnetRec.HisOpt[SGA].Accept := true;
  TelnetRec.HisOpt[ECHO].Accept := true;
  TelnetRec.MyOpt[TermType].Accept := true;
  TelnetRec.MyOpt[NAWS].Accept := true;
  TelnetRec.HisOpt[NAWS].Accept := true;
  TelnetRec.WinSize.X := ts.TerminalWidth;
  TelnetRec.WinSize.Y := ts.TerminalHeight;

  if ts.TelLog then
    TelnetRec.LogFile := _lcreat('TELNET.LOG',0)
  else
    TelnetRec.LogFile := 0;
end;


procedure DefaultTelRec;
var
  i: integer;
begin
with TelnetRec do begin

  for i:=0 to MaxTelOpt do
  begin
    MyOpt[i].Accept := false;
    MyOpt[i].Status := No;
    MyOpt[i].Que := Empty;
    HisOpt[i].Accept := false;
    HisOpt[i].Status := No;
    HisOpt[i].Que := Empty;
  end;

  SubOptCount := 0;
  SubOptIAC := false;
  ChangeWinSize := false;
end;
end;

procedure EndTelnet;
begin
  if TelnetRec.LogFile<>0 then
  begin
    TelnetRec.LogFile := 0;
    _lclose(TelnetRec.LogFile);
  end;
end;

procedure TelWriteLog1(b: byte);
var
  Temp: array[0..2] of char;
  Ch: byte;
begin
  Temp[0] := #$20;
  Ch := b div 16;
  case Ch of
    0..9:   Ch := Ch + $30;
    10..15: Ch := Ch + $37;
  end;
  Temp[1] := char(Ch);
  Ch := b and 15;
  case Ch of
    0..9:   Ch := Ch + $30;
    10..15: Ch := Ch + $37;
  end;
  Temp[2] := char(Ch);
  _lwrite(TelnetRec.LogFile,Temp,3);
end;

procedure TelWriteLog(Buf: PChar; C: integer);
var
  i: integer;
begin
  _lwrite(TelnetRec.LogFile,#$0d#$0a'>',3);
  for i := 0 to C-1 do
    TelWriteLog1(byte(Buf[i]));
end;

procedure SendBack(a, b: byte);
var
  Str3: array[0..2] of char;
begin
  Str3[0] := char(IAC);
  Str3[1] := char(a);
  Str3[2] := char(b);
  CommRawOut(@cv,Str3,3);
  if ts.TelLog then
    TelWriteLog(Str3,3);
end;

procedure ParseTel;

  procedure ParseTelIAC(b: byte);
  begin
  with TelnetRec do begin
    case b of
      SE:  ;
      NOP:     TelStatus := TelIdle;
      DM:      TelStatus := TelIdle;
      BREAK:   TelStatus := TelIdle;
      IP:      TelStatus := TelIdle;
      AO:      TelStatus := TelIdle;
      AYT:     TelStatus := TelIdle;
      EC:      TelStatus := TelIdle;
      EL:      TelStatus := TelIdle;
      GOAHEAD: TelStatus := TelIdle;
      SB:      begin
                 TelStatus := TelSB;
                 SubOptCount := 0;
               end;
      WILLTEL: TelStatus := TelWill;
      WONTTEL: TelStatus := TelWont;
      DOTEL:   TelStatus := TelDo;
      DONTTEL: TelStatus := TelDont;
      IAC: begin
             TelStatus := TelIdle;
           end;
    else
      TelStatus := TelIdle;
    end;
  end;
  end;

  procedure ParseTelSB(b: byte);
  var
    TmpStr: array[0..50] of char;
    i: integer;
  begin
  with TelnetRec do begin
    if SubOptIAC
    then begin
      SubOptIAC := false;
      case b of
        SE: begin
          if (MyOpt[TermType].Status = Yes) and
             (SubOptCount >= 2) and
             (SubOptBuff[0] = TermType) and
             (SubOptBuff[1] = 1) then
          begin
            TmpStr[0] := char(IAC);
            TmpStr[1] := char(SB);
            TmpStr[2] := char(TermType);
            TmpStr[3] := #0;
            StrCopy(@TmpStr[4],ts.TermType);
            i := 4 + StrLen(ts.TermType);
            TmpStr[i] := char(IAC);
            inc(i);
            TmpStr[i] := char(SE);
            inc(i);
            CommRawOut(@cv,TmpStr,i);

            if ts.TelLog then
              TelWriteLog(TmpStr,i);
          end
          else if {(HisOpt[NAWS].Status = Yes) and}
                  (SubOptCount >= 5) and
                  (SubOptBuff[0] = NAWS) then
          begin
            WinSize.X := SubOptBuff[1]*256+
                         SubOptBuff[2];
            WinSize.Y := SubOptBuff[3]*256+
                         SubOptBuff[4];
            ChangeWinSize := true;
          end;
          SubOptCount := 0;
          TelStatus := TelIdle;
          exit;
        end;
        {IAC: ;}
      else
        if SubOptCount >= SizeOf(SubOptBuff)-1 then
        begin
          SubOptCount := 0;
          TelStatus := TelIdle;
          exit;
        end
        else begin
          SubOptBuff[SubOptCount] := IAC;
          inc(SubOptCount);
          if b=IAC then
          begin
            SubOptIAC := true;
            exit;
          end;
        end;
      end;
    end
    else
      if b=IAC then
      begin
        SubOptIAC := true;
        exit;
      end;
 
    if SubOptCount >= SizeOf(SubOptBuff)-1 then
    begin
      SubOptCount := 0;
      SubOptIAC := false;
      TelStatus := TelIdle;
    end
    else begin
      SubOptBuff[SubOptCount] := b;
      inc(SubOptCount);
    end;
  end;
  end;

  procedure ParseTelWill(b: byte);
  begin
  with TelnetRec do begin
    if b <= MaxTelOpt then
      case HisOpt[b].Status of
        No: begin
             if HisOpt[b].Accept
             then begin
               SendBack(DOTEL,b);
               HisOpt[b].Status := Yes;
             end
             else
               SendBack(DONTTEL,b);
            end;
  
        WantNo: case HisOpt[b].Que of
                  Empty: HisOpt[b].Status := No;
                  Opposite: HisOpt[b].Status := Yes;
                end;

        WantYes: case HisOpt[b].Que of
                   Empty: HisOpt[b].Status := Yes;
                   Opposite: begin
                     HisOpt[b].Status := WantNo;
                     HisOpt[b].Que := Empty;
                     SendBack(DONTTEL,b);
                   end;
                 end;
      end
    else
      SendBack(DONTTEL,b);

    case b of
      BINARY:
        case HisOpt[BINARY].Status of
          Yes: cv.TelBinRecv := true;
          No:  cv.TelBinRecv := false;
        end;
    end;
    TelStatus := TelIdle;
  end;
  end;

  procedure ParseTelWont(b: byte);
  begin
  with TelnetRec do begin
    if b <= MaxTelOpt then
      case HisOpt[b].Status of
        Yes: begin
               HisOpt[b].Status := No;
               SendBack(DONTTEL,b);
             end;

        WantNo: case HisOpt[b].Que of
                  Empty: HisOpt[b].Status := No;
                  Opposite: begin
                    HisOpt[b].Status := WantYes;
                    HisOpt[b].Que := Empty;
                    SendBack(DOTEL,b);
                  end;
                end;

        WantYes: case HisOpt[b].Que of
                   Empty: HisOpt[b].Status := No;
                   Opposite: begin
                     HisOpt[b].Status := No;
                     HisOpt[b].Que := Empty;
                   end;
                 end;
      end
    else
      SendBack(DONTTEL,b);

    case b of
      BINARY:
        case HisOpt[BINARY].Status of
          Yes: cv.TelBinRecv := true;
          No:  cv.TelBinRecv := false;
        end;
    end;
    TelStatus := TelIdle;
  end;
  end;

  procedure ParseTelDo(b: byte);
  begin
  with TelnetRec do begin
    if b <= MaxTelOpt then
    begin
      case MyOpt[b].Status of
        No: begin
             if MyOpt[b].Accept
             then begin
               MyOpt[b].Status := Yes;
               SendBack(WILLTEL,b);
             end
             else
               SendBack(WONTTEL,b);
            end;

        WantNo: case MyOpt[b].Que of
                  Empty: MyOpt[b].Status := No;
                  Opposite: MyOpt[b].Status := Yes;
                end;

        WantYes: case MyOpt[b].Que of
                   Empty: MyOpt[b].Status := Yes;
                   Opposite: begin
                     MyOpt[b].Status := WantNo;
                     MyOpt[b].Que := Empty;
                     SendBack(WONTTEL,b);
                   end;
                 end;
      end;
    end
    else
      SendBack(WONTTEL,b);

    case b of
      BINARY:
        case MyOpt[BINARY].Status of
          Yes: cv.TelBinSend := true;
          No:  cv.TelBinSend := false;
        end;
      NAWS: if MyOpt[NAWS].Status=Yes then
          TelInformWinSize;
    end;
    TelStatus := TelIdle;
  end;
  end;

  procedure ParseTelDont(b: byte);
  begin
  with TelnetRec do begin
    if b <= MaxTelOpt then
      case MyOpt[b].Status of
        Yes: begin
               MyOpt[b].Status := No;
               SendBack(WONTTEL,b);
             end;
   
        WantNo: case MyOpt[b].Que of
                  Empty: MyOpt[b].Status := No;
                  Opposite: begin
                    MyOpt[b].Status := WantYes;
                    MyOpt[b].Que := Empty;
                    SendBack(WILLTEL,b);
                  end;
                end;  

        WantYes: case MyOpt[b].Que of
                   Empty: MyOpt[b].Status := No;
                   Opposite: begin
                     MyOpt[b].Status := No;
                     MyOpt[b].Que := Empty;
                   end;
                 end;
      end
    else
      SendBack(WONTTEL,b);

    case b of
      BINARY:
        case MyOpt[BINARY].Status of
          Yes: cv.TelBinSend := true;
          No:  cv.TelBinSend := false;
        end;
    end;
    TelStatus := TelIdle;
  end;
  end;


var
  b: byte;
  c: integer;
begin
with TelnetRec do begin

  c := CommReadRawByte(@cv,@b);

  while (c>0) and (cv.TelMode) do
  begin
    if ts.TelLog then
    begin
      if TelStatus=TelIAC then
      begin
        _lwrite(LogFile,#$0d#$0a'<',3);
        TelWriteLog1($ff);
      end;
      TelWriteLog1(b);
    end;

    ChangeWinSize := false;

    case TelStatus of
      TelIAC: ParseTelIAC(b);
      TelSB: ParseTelSB(b);
      TelWill: ParseTelWill(b);
      TelWont: ParseTelWont(b);
      TelDo: ParseTelDo(b);
      TelDont: ParseTelDont(b);
      TelNop: TelStatus := TelIdle;
    end;
    if TelStatus = TelIdle then cv.TelMode := false;

    if cv.TelMode then c := CommReadRawByte(@cv,@b);
  end;

end;
end;


procedure TelEnableHisOpt(b: byte);
begin
with TelnetRec do begin
  if b <= MaxTelOpt then
  case HisOpt[b].Status of
    No: begin
          HisOpt[b].Status := WantYes;
          SendBack(DOTEL,b);
        end;

    WantNo: if HisOpt[b].Que=Empty then
               HisOpt[b].Que := Opposite;

    WantYes: if HisOpt[b].Que=Opposite then
                HisOpt[b].Que := Empty;
  end;                              
end;
end;

procedure TelDisableHisOpt(b: byte);
begin
with TelnetRec do begin
  if b <= MaxTelOpt then
  case HisOpt[b].Status of
    Yes: begin
           HisOpt[b].Status := WantNo;
           SendBack(DONTTEL,b);
         end;

    WantNo: if HisOpt[b].Que=Opposite then
               HisOpt[b].Que := Empty;

    WantYes: if HisOpt[b].Que=Empty then
                HisOpt[b].Que := Opposite;
  end;
end;
end;

procedure TelEnableMyOpt(b: byte);
begin
with TelnetRec do begin
  if b <= MaxTelOpt then
  case MyOpt[b].Status of
    No: begin
          MyOpt[b].Status := WantYes;
          SendBack(WILLTEL,b);
        end;

    WantNo: if MyOpt[b].Que=Empty then
               MyOpt[b].Que := Opposite;

    WantYes: if MyOpt[b].Que=Opposite then
                MyOpt[b].Que := Empty;
  end;
end;
end;

procedure TelDisableMyOpt(b: byte);
begin
with TelnetRec do begin
  if b <= MaxTelOpt then
  case MyOpt[b].Status of
    Yes: begin
           MyOpt[b].Status := WantNo;
           SendBack(WONTTEL,b);
         end;

    WantNo: if MyOpt[b].Que=Opposite then
               MyOpt[b].Que := Empty;

    WantYes: if MyOpt[b].Que=Empty then
                MyOpt[b].Que := Opposite;
  end;
end;
end;

procedure TelInformWinSize;
var
  i: integer;
  TmpBuff: array[0..20] of char;
begin
with TelnetRec do begin
  i := 0;

  TmpBuff[i] := char(IAC);
  Inc(i);
  TmpBuff[i] := char(SB);
  Inc(i);
  TmpBuff[i] := char(NAWS);
  Inc(i);
  TmpBuff[i] := char(Hi(WinSize.X));
  Inc(i);
 { if Lo(WinSize.X) = IAC then
  begin
    SendBackBuff[i] := IAC;
    Inc(i);
  end; }
  TmpBuff[i] := char(Lo(WinSize.X));
  Inc(i);
  TmpBuff[i] := char(Hi(WinSize.Y));
  Inc(i);
 { if Lo(WinSize.Y) = IAC then
  begin
    SendBackBuff[i] := IAC;
    Inc(i);
  end; }
  TmpBuff[i] := char(Lo(WinSize.Y));
  Inc(i);
  TmpBuff[i] := char(IAC);
  Inc(i);
  TmpBuff[i]:= char(SE);
  Inc(i);

  CommRawOut(@cv,TmpBuff,i);
  if ts.TelLog then
    TelWriteLog(TmpBuff,i);
end;
end;

procedure TelSendAYT;
var
  Str: array[0..1] of char;
begin
  Str[0] := char(IAC);
  Str[1] := char(AYT);
  CommRawOut(@cv,Str,2);
  CommSend(@cv);
  if ts.TelLog then
    TelWriteLog(Str,2);
end;

procedure TelSendBreak;
var
  Str: array[0..1] of char;
begin
  Str[0] := char(IAC);
  Str[1] := char(BREAK);
  CommRawOut(@cv,Str,2);
  CommSend(@cv);
  if ts.TelLog then
    TelWriteLog(Str,2);
end;

end.